
'''
	import fried_tofu.node.stop as node_stop
	node_stop.attractively (
		process_ID_file_path,
		driver_port
	)
		
'''

'''
	notes:
		rethinkdb removes the pid file when stopped
		like this.
'''

import fried_tofu.node.cannot_connect as node_cannot_connect

import psutil

def attractively (
	process_ID_file_path,
	driver_port
):
	print ()
	print ("[stopping rethink node]")

	FP = open (process_ID_file_path)
	P_ID = int (FP.read ().strip ())

	p = psutil.Process (P_ID)
	p.terminate ()
	
	node_cannot_connect.attractively (
		loops = 2,
		driver_port = driver_port
	)
	
	return;




'''
	import fried_tofu.revenue as fried_tofu_revenue
	fried_tofu_revenue.obtain ()
'''


import pathlib
from os.path import dirname, join, normpath
this_folder = pathlib.Path (__file__).parent.resolve ()

import revenue


def obtain ():
	revenue.start ({
		"directory": str (this_folder),
		"extension": ".r.HTML",
		"relative path": str (this_folder)
	})





'''
import fried_tofu.node.cannot_connect as fried_tofu_cannot_connect
fried_tofu_cannot_connect.attractively (
	loops = 3,
	driver_port = driver_port
)
'''

import fried_tofu.node.connect as connect

def attractively (
	loops = 5,
	driver_port = None
):
	int (driver_port)

	connect.attractively (
		label = "asserting that can't connect",
		driver_port = driver_port,
		loops = loops,
		assert_cannot_connect = True
	)











'''
	priorities:
		import pathlib
		from os.path import dirname, join, normpath
		this_folder = pathlib.Path (__file__).parent.resolve ()
		
		import fried_tofu.node.process_IDs.remove as remove_process_IDs
		remove_process_IDs.attractively (
			glob_string = str (normpath (join (this_folder, "assurances"))) + "**/tofu.process_ID"
		)
'''

import glob
import os

def attractively (
	glob_string
):
	paths_found = glob.glob (glob_string)

	for path_found in paths_found:
		if (os.path.isfile (path_found)):
			os.remove (path_found)

	return;




'''
	# for status checks

	import pathlib
	from os.path import dirname, join, normpath
	this_directory = pathlib.Path (__file__).parent.resolve ()	
	cluster_directory = normpath (join (this_directory, "tofu_cluster"))
	import fried_tofu.cluster.start as start_cluster
	ly = start_cluster.attractively (
		proxy = {
			"ports": []
		},
		nodes = [{
			"ports": []
		},{
			"ports": []
		},{
			"ports": []
		}]
	)
'''

'''	
	import fried_tofu.cluster.start as start_cluster
	ly = start_cluster.attractively (
		proxy = {
			"ports": []
		},
		nodes = [{
			"ports": []
		},{
			"ports": []
		},{
			"ports": []
		}]
	)
'''

'''
	rethinkdb proxy --join 127.0.0.1:29016 --daemon
	
	rethinkdb -o 1 --bind all --daemon
		# Listening for intracluster connections on port 29016
		# Listening for client driver connections on port 28016
		# Listening for administrative HTTP connections on port 8081
	
	rethinkdb -o 2 --join 127.0.0.1:29016 --daemon
	rethinkdb -o 3 --join 127.0.0.1:29016 --daemon
'''

'''
	"join" needs one of the hosts allowed with "--bind", etc.
	
		Listening on cluster addresses: 127.0.0.1, 192.168.0.11, ::1, fe80::b3d:4b55:4fab:7aa1%2
		Listening on driver addresses: 127.0.0.1, 192.168.0.11, ::1, fe80::b3d:4b55:4fab:7aa1%2
		Listening on http addresses: 127.0.0.1, 192.168.0.11, ::1, fe80::b3d:4b55:4fab:7aa1%2
'''

import pathlib
from os.path import dirname, join, normpath
import os

import botany
import botany.flow.demux_mux2 as demux_mux2

import fried_tofu.node.stop as node_stop
import fried_tofu.node.start as start_node



import inspect
import os

def retrieve_caller ():
	return os.path.abspath ((inspect.stack () [1]) [1])

def attractively (
	proxy = {
		"ports": []
	},
	nodes = [],
	
	#
	#	tofu_cluster in caller directory if not sent.
	#
	cluster_directory = "",
	process_ID_file_name = "tofu.pid"
):
	shutdown_list = []
	
	if (len (cluster_directory) == 0):
		cluster_directory = str (normpath (join (
			os.path.dirname (os.path.abspath ((inspect.stack () [1]) [1])), 
			"tofu_cluster"
		)))

	print ("cd", cluster_directory);

	node_tally = 1
	def start (parameters):
		nonlocal node_tally
		nonlocal cluster_directory

		node_path = str (normpath (join (cluster_directory, str (node_tally))))
		process_ID_file_path = str (normpath (join (node_path, process_ID_file_name)))
		os.makedirs (node_path, exist_ok = True)
		
		ports = parameters ["ports"]
		driver_port = ports [0]

		ly = start_node.attractively (
			proxy = "no",
			server_name = f"node-{ node_tally }",
			ports = {
				"driver": driver_port,
				"cluster": ports [1],
				"http": ports [2]	
			},
			process = {
				"cwd": node_path
			},
			rethink_params = [
				f"--daemon",
				f"--pid-file { process_ID_file_path }"
			]
		)
		
		shutdown_list.append ({
			"process_ID_file_path": process_ID_file_path,
			"driver_port": driver_port
		})
		
		node_tally += 1

	proceeds_statement = demux_mux2.start (
		start, 
		nodes
	)

	return;





















'''
	import pathlib
	import fried_tofu.node.start as start_node
	ly = start_node.attractively (
		proxy = "yes",
		server_name = "node-1",
		ports = {
			"driver": 0,
			"cluster": 0,
			"http": 0	
		},
		process = {
			"cwd": pathlib.Path (__file__).parent.resolve ()
		},
		rethink_params = [
			f"--daemon",
			f"--pid-file {}"
		]
	)
'''

import shlex
import subprocess
import time
import datetime
	
import fried_tofu.node.start.assertions as node_start_assertions
import fried_tofu.node.connect as connect_to_node
	
def attractively (
	proxy = "no",
	
	rethink_params = [],
	ports = {},
	
	records = 1,
	
	** keywords
):
	print ()
	print ("[starting rethink node]")

	#
	process_keys = keywords ["process"]

	#
	driver_port = str (ports ["driver"])
	cluster_port = str (ports ["cluster"])
	http_port = str (ports ["http"])
	
	node_string = "rethinkdb"
	if (proxy == "yes"):
		#
		#	proxies can't have names maybe.
		#
		node_string += " proxy"
	
	'''
	#timestamp = datetime.datetime ().timestamp()
	timestamp = ''.join (str (datetime.datetime.now().timestamp()).split ('.'))
	print ('timestamp:', timestamp)
	f"--server-tag { timestamp }",
	'''
	
	script = " ".join ([
		node_string,
		
		f"--driver-port { driver_port }",
		f"--cluster-port { cluster_port }",
		f"--http-port { http_port }",
		
		* rethink_params
	])
	
	print ("script:", script)
	
	process = subprocess.Popen (
		shlex.split (script),
		** process_keys
	)
	
	time.sleep (1)
	
	[ r, c ] = connect_to_node.attractively (
		driver_port = driver_port,
		loops = 5,
		delay = 1
	)
	
	'''
	assert (r.db('rethinkdb').table('server_config').count ().run (c) == 1)
	server_config = r.db('rethinkdb').table('server_config').run (c)
	for config in server_config:
		assert (timestamp in config ["tags"]), [
			config,
			timestamp
		]
	'''
	
	'''
	server_data = c.server ();
	print ("server_data:", server_data)
	assert (server_data ["name"] == server_name)
	'''
	
	node_start_assertions.check (r, c)		
	c.close ()

	return;







'''
	import fried_tofu.node.connect as connect_to_node
	[ r, c ] = connect_to_node.attractively (
		driver_port = "",
		loops = 3,
		delay = 2
	)
'''


from rethinkdb import RethinkDB

import botany.cycle as cycle
import botany.modules.exceptions.parse as parse_exception

class connection_exception (Exception):
	pass


def attractively (
	loops = 10,
	delay = 1,
	
	driver_port = None,
	host = "localhost",
	
	label = "connect",
	
	assert_cannot_connect = False
):
	if (driver_port == None):
		ports = climate.find ("ports")
		driver_port = ports ["driver"]

	connection_attempt = 1;
	def connect (* positionals, ** keywords):	
		nonlocal connection_attempt;
		print (
			f"{ label }: Attempt '{ connection_attempt }' to connect to rethink at: { host }:{ driver_port }", 	
		)
		
		connection_attempt += 1
		
		r = RethinkDB ()
		c = r.connect (
			host = host,
			port = driver_port
		)

		print (f'Connection to rethink at "{ host }:{ driver_port }" was established.');	

		return [ r, c ];
		
	connection_loop_exception_occurred = False
	try:
		connection = cycle.loops (
			connect, 
			cycle.presents ([]),
			
			loops = loops,
			delay = delay,
			
			records = 0
		)
		
	except Exception as E:
		if (assert_cannot_connect):
			connection_loop_exception_occurred = True
			
		else:
			parsed_exception = parse_exception.now (E)
			print ("connection loop exception:", str (E))		
			raise connection_exception (f'''
			
				"{ loops }" connection attempts were made,
				however a connection could not be established.
				
			''')
		
	if (assert_cannot_connect):
		assert (connection_loop_exception_occurred == True)
		return;
	
	return connection;
